TC_BANDWIDTH_UNITS = [
    "kbit",
    "mbit",
    "gbit",
    "tbit",
    "kbps",
    "mbps",
    "gbps",
    "tbps",
    # keep the shorter units at the end
    "bit",
    "bps",
]


def split_raw_str_rate(rate: str) -> tuple[int, str]:
    """Split raw string rate into rate and unit.
    Args:
        - rate (str) : raw string rate. e.g. "100mbit"

    Returns:
        - rate (int) : rate
        - unit (str) : rate unit
    """
    rate = rate.lower()
    for unit in TC_BANDWIDTH_UNITS:
        if rate.endswith(unit):
            return int(rate[: -len(unit)]), unit
    raise ValueError(
        f"Invalid rate {rate}. " + f"Valid units are {TC_BANDWIDTH_UNITS}"
    )



import socket
import os
import struct
import pickle
import time
import fcntl
import errno
import asyncio
from asyncio import Queue
from typing import Optional

from .msg import CtrlMsg, CtrlAction


class NetCtrlCommServer:
    """Network control communication server.
    This class is used to create a unix socket server
    to receive control messages from clients.
    Typically, only one client will connect to the server.
    Args:
        - socket_path (str) : path to the unix socket
    """

    def __init__(self, socket_path: str):
        self.socket_path = socket_path
        self._lock_file = None
        self._check_path()
        self._server: Optional[asyncio.AbstractServer] = None
        self._stop_event = asyncio.Event()

    async def start(self, q: Queue):
        """Start the server. This method will start the server
        and wait for incoming messages, and put
        them in the queue.
        Args:
            - q (Queue) : asyncio.Queue object to put incoming messages
        """
        try:
            self._server = await asyncio.start_unix_server(
                lambda reader, writer: self._handle_client(reader, writer, q),
                path=self.socket_path,
            )
        except OSError as e:
            if e.errno in (errno.EADDRINUSE, errno.EACCES):
                raise OSError(
                    f"Failed to bind to socket: {self.socket_path}"
                ) from e
            else:
                raise
        try:
            serve_task = asyncio.create_task(self._server.serve_forever())
            await self._stop_event.wait()  # stop event
            serve_task.cancel()  # cancel the serve task
            try:
                await serve_task
            except asyncio.CancelledError:
                pass
        finally:
            await self.stop()

    async def _recvall(self, reader, n):
        """Receive n bytes from the reader.
        Args:
            - reader : asyncio.StreamReader object
            - n (int) : number of bytes to receive
        """
        data = bytearray()
        while len(data) < n:
            packet = await reader.read(n - len(data))
            if not packet:
                return None
            data.extend(packet)
        return data

    def _check_path(self):
        """Check if the socket path exists and is in use."""
        lock_file = self.socket_path + ".lock"
        self._lock_file = open(lock_file, "w")

        try:
            fcntl.flock(self._lock_file.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
        except BlockingIOError:
            raise OSError(
                "Another instance is holding the lock. "
                f"Socket path: {self.socket_path}"
            )

        if os.path.exists(self.socket_path):
            try:
                # Attempt to connect to the existing socket
                with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as s:
                    s.settimeout(1)  # set timeout to 1 second
                    s.connect(self.socket_path)
                # If the connection is successful, the socket is in use
                raise OSError(
                    f"Socket is already in use. Socket path: {self.socket_path}"
                )
            except OSError as e:
                if e.errno == errno.ECONNREFUSED:
                    # If the connection is refused, the socket is not in use
                    os.remove(self.socket_path)
                elif e.errno == errno.ENOENT:
                    # If the socket does not exist, it is not in use
                    pass
                else:
                    raise

    async def _handle_client(self, reader, writer, q: Queue):
        """Handle the client connection."""
        try:
            while True:
                raw_msglen = await self._recvall(reader, 4)
                if not raw_msglen:
                    break
                msglen = struct.unpack(">I", raw_msglen)[0]
                data = await self._recvall(reader, msglen)
                if not data:
                    break
                data_obj = pickle.loads(data)
                # put the data in the queue
                await q.put(data_obj)
                if data_obj.action == CtrlAction.STOP:
                    asyncio.create_task(self._initiate_stop())
                    break
        finally:
            # close the writer
            writer.close()
            await writer.wait_closed()

    async def _initiate_stop(self):
        """Initiate the stop event."""

        self._stop_event.set()

    async def stop(self):
        """Stop the server."""
        if self._server:
            self._server.close()
            await self._server.wait_closed()
        self._stop_event.set()


class NetCtrlCommClient:
    def __init__(
        self, socket_path: str, _conn_retry: int = 5, _retry_interval: int = 1
    ):
        self.socket_path = socket_path
        self.conn: Optional[socket.socket] = None
        self._conn_retry = _conn_retry
        self._retry_interval = _retry_interval
        self._connected = False
        self._connect()

    def _connect(self):
        self._check_path()
        for _ in range(self._conn_retry):
            try:
                self.conn = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                self.conn.connect(self.socket_path)
                break
            except ConnectionRefusedError:
                print(
                    f"Connection refused. Retrying \
                    in {self._retry_interval} seconds."
                )
                time.sleep(self._retry_interval)
        else:
            raise ConnectionRefusedError(
                f"Cound not connect to {self.socket_path}"
            )
        self._connected = True

    def send(self, data: CtrlMsg):
        """Send data to the server."""
        if not self._connected:
            raise ConnectionError("Connection has been closed.")

        serialized_data = pickle.dumps(data)
        data_len = struct.pack(">I", len(serialized_data))
        if self.conn is not None:
            self.conn.sendall(data_len + serialized_data)
        else:
            raise ConnectionError("Connection is not established.")

    def stop_server(self):
        """Stop the server. This method will send a stop message
        to the server and close the connection.
        """
        self.send(CtrlMsg(action=CtrlAction.STOP))  # send stop message
        self.stop()

    def stop(self):
        """Stop the client."""
        self.conn.close()
        self._connected = False

    def _check_path(self):
        """Check if the socket path exists, and remove it if it does."""
        if not os.path.exists(self.socket_path):
            raise FileNotFoundError(
                f"Socket file not found: {self.socket_path}"
            )



from enum import Enum
from dataclasses import dataclass
from typing import Optional


class CtrlAction(Enum):
    SET_BANDWIDTH = 1  # set bandwidth limit between two containers
    ADD_CONTAINER = 2  # add a new container to monitor
    DEL_CONTAINER = 3  # remove a container from monitoring
    STOP = 4  # stop monitoring and adjusting network


@dataclass
class CtrlMsg:
    action: CtrlAction
    container: Optional[str] = None



from .cmd_wrapper import TCCmdWrapper, DockerCmdWrapper
from .container_net_ctrl import ConNetController, ConNetServer

__version__ = "0.0.4"

__all__ = [
    "TCCmdWrapper",
    "DockerCmdWrapper",
    "ConNetController",
    "ConNetServer",
]



import argparse
import sys
import os
import contcfg
from contcfg import ConNetServer, ConNetController, TCCmdWrapper
from contcfg.cmd_wrapper import split_raw_str_rate

__version__ = contcfg.__version__


def is_running_as_root():
    return os.geteuid() == 0


def handle_exception(func):
    """Decorator to handle exceptions for server operations."""

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ConnectionRefusedError:
            print("Connection refused. Server may not be running.")
        except Exception as e:
            print(f"Error: {e}")
            sys.exit(1)

    return wrapper


@handle_exception
def start_server(args, run_with_sudo):
    min_rate, min_unit = split_raw_str_rate(args.min_rate)
    max_rate, max_unit = split_raw_str_rate(args.max_rate)
    if min_unit != max_unit:
        raise ValueError("Rate units do not match. Please provide same units.")
    server = ConNetServer(
        min_rate,
        max_rate,
        args.interval,
        rate_unit=min_unit,
        interval_unit="min",
        _server_socket_path=args.socket_path,
        _run_with_sudo=run_with_sudo,
    )
    server.start()


@handle_exception
def ctrl(args, run_with_sudo):
    sender = ConNetController(
        _socket_path=args.socket_path, _run_with_sudo=run_with_sudo
    )
    if args.ctrl_command == "add":
        sender.add_container(args.container)
    elif args.ctrl_command == "del":
        sender.del_container(args.container)
    elif args.ctrl_command == "add-all":
        sender.add_all_containers(args.prefix)
    sender.stop()


@handle_exception
def stop_server(args, run_with_sudo):
    sender = ConNetController(
        _socket_path=args.socket_path, _run_with_sudo=run_with_sudo
    )
    sender.stop_server()


@handle_exception
def run_cli(args, run_with_sudo):
    if args.cli_command == "set":
        TCCmdWrapper(run_with_sudo).set_bandwidth(
            args.container1, args.container2, args.bandwidth
        )
    elif args.cli_command == "clear":
        for c in args.containers:
            TCCmdWrapper(run_with_sudo).clear_one_container(c)


def create_parser():
    """Creates and returns the argument parser."""
    parser = argparse.ArgumentParser(description="Container Network Controller")
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + __version__
    )
    parser.add_argument(
        "--socket-path",
        type=str,
        default="/tmp/contcfg.sock",
        help="Path to the Unix socket",
    )

    subparsers = parser.add_subparsers(dest="command", help="Sub-command help")

    # Start server sub-command
    server_parser = subparsers.add_parser(
        "start-server", help="Start the server"
    )
    server_parser.add_argument("min_rate", type=str, help="Minimum rate")
    server_parser.add_argument("max_rate", type=str, help="Maximum rate")
    server_parser.add_argument("interval", type=int, help="Interval in minutes")

    # Control server sub-command
    ctrl_parser = subparsers.add_parser("ctrl", help="Control the server")
    ctrl_subparsers = ctrl_parser.add_subparsers(
        dest="ctrl_command", help="Control sub-command help"
    )
    # Add and delete container sub-commands
    for action in ["add", "del"]:
        sub_parser = ctrl_subparsers.add_parser(
            action, help=f"{action.capitalize()} container"
        )
        sub_parser.add_argument(
            "container", type=str, help="Container name or id"
        )
    # Add all containers sub-command
    sub_parser = ctrl_subparsers.add_parser(
        "add-all", help="Add all containers"
    )
    sub_parser.add_argument("prefix", type=str, help="Container name prefix")

    # Stop server sub-command
    subparsers.add_parser("stop-server", help="Stop the server")

    # Cli of wrapper
    cli_parser = subparsers.add_parser("cli", help="Run the CLI")

    cli_subparsers = cli_parser.add_subparsers(
        dest="cli_command", help="CLI sub-command help"
    )
    # Set bandwidth sub-command
    sub_parser = cli_subparsers.add_parser("set", help="Set bandwidth")

    sub_parser.add_argument("container1", type=str, help="Container name or id")
    sub_parser.add_argument("container2", type=str, help="Container name or id")
    sub_parser.add_argument("bandwidth", type=str, help="Bandwidth limit")

    # Clear bandwidth sub-command
    sub_parser = cli_subparsers.add_parser("clear", help="Clear bandwidth")
    sub_parser.add_argument(
        "containers", type=str, nargs="+", help="Container name(s) or id(s)"
    )

    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()
    run_with_sudo = is_running_as_root()
    if args.command == "start-server":
        start_server(args, run_with_sudo)
    elif args.command == "ctrl":
        ctrl(args, run_with_sudo)
    elif args.command == "stop-server":
        stop_server(args, run_with_sudo)
    elif args.command == "cli":
        run_cli(args, run_with_sudo)
    else:
        parser.print_help()
        sys.exit(1)


if __name__ == "__main__":
    main()



import random
import asyncio
from asyncio import Queue
import logging
from concurrent.futures import ThreadPoolExecutor
from typing import Optional


from ..cmd_wrapper import (
    TCCmdWrapper,
    DockerCmdWrapper,
    RateValueError,
    ContainerNotFoundError,
)
from .msg import CtrlMsg, CtrlAction
from .comm import NetCtrlCommServer

__all__ = ["ConNetServer"]

logging.basicConfig(
    level=logging.INFO, format="[%(asctime)s][%(levelname)s]%(message)s"
)


def all_pairs_iter(lst: list, paticular=None):
    """Generate all pairs from a list. If paticular is given,
    generate pairs including paticular."""
    if paticular:
        for i in range(len(lst)):
            if lst[i] != paticular:
                yield lst[i], paticular
    else:
        for i in range(len(lst)):
            for j in range(i + 1, len(lst)):
                yield lst[i], lst[j]


class ConNetServer:
    """Container network controller class.
    This class is used to control network bandwidth
    between docker containers.
    Args:
        - min_rate (int) : minimum rate in mbit
        - max_rate (int) : maximum rate in mbit
        - interval (int) : interval in seconds
        - prefix (str, optional) : container name prefix.
        - _server_socket_path (str, optional) : path to the Unix socket.
        - _run_with_sudo (bool, optional) : run all commands with sudo.
    Kwargs:
        - rate_unit (str, optional) : rate unit. Default is "mbit".
        - interval_unit (str, optional) : interval unit. Default is "min".
        Default is None.
        - _run_with_sudo (bool, optional) : run all commands with sudo.
        Default is False.
    """

    def __init__(
        self,
        min_rate: int,
        max_rate: int,
        interval: int,
        *,
        prefix: str = "",
        _server_socket_path: str = "/tmp/contcfg.sock",
        _run_with_sudo: bool = False,
        **kwargs,
    ):
        self.min_rate = min_rate
        self.max_rate = max_rate
        self.rate_unit = kwargs.get("rate_unit", "mbit")
        self.interval_unit = kwargs.get("interval_unit", "min")
        self.prefix = prefix
        self._run_with_sudo = _run_with_sudo
        # convert interval to seconds
        if self.interval_unit.startswith("s"):
            self.interval_sec = interval
        elif self.interval_unit.startswith("m"):
            self.interval_sec = interval * 60
        elif self.interval_unit.startswith("h"):
            self.interval_sec = interval * 3600

        self._container_list: list[str] = []
        self._msg_queue: Queue = asyncio.Queue()
        self._is_running = False
        self._limit_dict: dict[tuple[str, str], int] = {}
        self._stop_event = asyncio.Event()
        self._loop = asyncio.get_event_loop()
        self._socket_path = _server_socket_path

    async def _monitor_and_adjust_network(self):
        """Monitor and adjust network bandwidth between containers."""
        with ThreadPoolExecutor() as executor:
            while True:
                msg = await self._msg_queue.get()
                logging.debug(f"Received message: {msg}")
                if msg.action == CtrlAction.STOP:
                    # clear all bandwidth limits
                    tasks = []
                    for container in self._container_list:
                        tasks.append(
                            self._loop.run_in_executor(
                                executor, self._clear_one, container, True
                            )
                        )
                    # wait for all tasks to complete
                    await asyncio.gather(*tasks)
                    # set stop event
                    self._is_running = False
                    self._stop_event.set()
                    break
                elif msg.action == CtrlAction.SET_BANDWIDTH:
                    # for each container pair, set bandwidth limit
                    tasks = []
                    for container1, container2 in all_pairs_iter(
                        self._container_list
                    ):
                        tasks.append(
                            self._loop.run_in_executor(
                                executor,
                                self._set_bandwidth_limit,
                                container1,
                                container2,
                            )
                        )
                    # wait for all tasks to complete
                    await asyncio.gather(*tasks)
                elif msg.action == CtrlAction.ADD_CONTAINER:
                    # adjust network for new container
                    if msg.container in self._container_list:
                        continue

                    # initialize htb qdisc for the container. This is required
                    # because executor may call tc command
                    # in one container at a time,
                    # which may cause setting failed.
                    self._init_container_htb(msg.container)

                    tasks = []
                    for container1, container2 in all_pairs_iter(
                        self._container_list, msg.container
                    ):
                        tasks.append(
                            self._loop.run_in_executor(
                                executor,
                                self._set_bandwidth_limit,
                                container1,
                                container2,
                            )
                        )
                    # wait for all tasks to complete
                    await asyncio.gather(*tasks)
                    self._container_list.append(msg.container)
                elif msg.action == CtrlAction.DEL_CONTAINER:
                    if msg.container not in self._container_list:
                        logging.warning(
                            f"Recv DEL_CONTAINER. the container {msg.container}"
                            + " is not in the list"
                        )
                        continue
                    # clear bandwidth limit for the container
                    await self._loop.run_in_executor(
                        executor, self._clear_one, msg.container, True
                    )
                    self._container_list.remove(msg.container)
                # show bandwidth limits
                self._show_bandwidth_limits()

    async def _periodic_clock(self):
        """Periodic clock to trigger network adjustment."""
        while self._is_running:
            try:
                await asyncio.wait_for(
                    self._stop_event.wait(), timeout=self.interval_sec
                )
            except asyncio.TimeoutError:
                await self._msg_queue.put(CtrlMsg(CtrlAction.SET_BANDWIDTH))
        # stop the loop
        self._loop.stop()

    def start(self):
        """start monitoring and adjusting network."""
        server = NetCtrlCommServer(self._socket_path)
        self._is_running = True

        self._loop.create_task(self._monitor_and_adjust_network())
        self._loop.create_task(self._periodic_clock())
        self._loop.create_task(server.start(self._msg_queue))
        self._loop.run_forever()

    def _show_bandwidth_limits(self):
        """Show bandwidth limits between containers."""
        s = ""
        for (container1, container2), bandwidth in self._limit_dict.items():
            s += (
                f"{container1} <--> {container2} : {bandwidth} {self.rate_unit}"
            )
            s += "\n"
        if s:
            logging.info(f"Bandwidth limits:\n{s}")

    def _init_container_htb(self, container: str):
        """Initialize htb qdisc for container."""
        try:
            TCCmdWrapper(self._run_with_sudo).init_htb(container)
        except ContainerNotFoundError:
            logging.error(f"Container {container} not found")
        except Exception as e:
            logging.error(
                "Error initializing htb. Please check if tc is installed "
                f"or set docker run with --cap-add=NET_ADMIN: {e}"
            )

    def _set_bandwidth_limit(
        self, container1: str, container2: str, bandwidth: Optional[int] = None
    ):
        """Set bandwidth limit between two containers.
        Args:
            - container1 (str) : container name or id
            - container2 (str) : container name or id
            - bandwidth (int, optional) : bandwidth limit. Default is None.
        """
        if bandwidth is None:
            bandwidth = random.randint(self.min_rate, self.max_rate)

        try:
            container1, container2 = sorted([container1, container2])
            TCCmdWrapper(self._run_with_sudo).set_bandwidth(
                container1, container2, bandwidth, self.rate_unit
            )
            self._limit_dict[(container1, container2)] = bandwidth
        except ContainerNotFoundError:
            logging.error(f"Container {container1} or {container2} not found")
        except RateValueError as e:
            logging.error(f"Rate value error: {e}")
        except Exception as e:
            logging.error(
                "Error setting bandwidth. Please check if tc is installed"
                + f" or run with sudo: {e}"
            )

    def _clear_one(self, container: str, check_exist: bool = False):
        """Clear tc rules for one container."""
        if check_exist:
            if not DockerCmdWrapper(self._run_with_sudo).is_container_exist(
                container
            ):
                return
        try:
            TCCmdWrapper(self._run_with_sudo).clear_one_container(container)
        except ContainerNotFoundError:
            logging.error(f"Container {container} not found")
        except Exception as e:
            logging.error(f"Error clearing bandwidth limit: {e}")



from .comm import NetCtrlCommClient
from .msg import CtrlMsg, CtrlAction
from ..cmd_wrapper import DockerCmdWrapper


class ConNetController:
    """Container network controller message sender.
    This class is used to send control messages
    to the network controller.
    Args:
        - socket_path (str) : path to the unix socket
    """

    def __init__(
        self,
        prefix: str = "",
        *,
        _socket_path: str = "/tmp/contcfg.sock",
        _run_with_sudo: bool = False,
    ):
        self._socket_path = _socket_path
        self._client = NetCtrlCommClient(self._socket_path)
        self._run_with_sudo = _run_with_sudo
        self._prefix = prefix
        self._containers: list[str] = []

    def add_container(self, container: str):
        """Add container to the network controller.
        Args:
            - container (str) : container name or id
        """
        if container in self._containers:
            raise ValueError(f"Container {container} already exists.")
        try:
            DockerCmdWrapper(self._run_with_sudo).check_container(container)
        except Exception as e:
            raise e
        self._client.send(CtrlMsg(CtrlAction.ADD_CONTAINER, container))
        self._containers.append(container)

    def del_container(self, container: str):
        """Delete container from the network controller.
        Args:
            - container (str) : container name or id
        """
        if container not in self._containers:
            raise ValueError(f"Container {container} not found.")
        self._client.send(CtrlMsg(CtrlAction.DEL_CONTAINER, container))
        self._containers.remove(container)

    def find_all_containers(self, prefix: str = ""):
        """Find all containers with given prefix."""
        prefix = prefix if prefix else self._prefix
        if not prefix:
            raise ValueError("Prefix is not set.")

        containers = DockerCmdWrapper(self._run_with_sudo).get_container(prefix)
        return containers

    def add_all_containers(self, prefix: str):
        """Add all containers with given prefix."""
        containers = self.find_all_containers(prefix)
        for container in containers:
            if container not in self._containers:
                self.add_container(container)

    def stop_server(self):
        """Stop the server."""
        self._client.stop_server()

    def stop(self):
        """Stop the client."""
        self._client.stop()



import subprocess
from typing import Optional

from .base import singleton, exec_cmd
from .exception import ContainerNotFoundError


@singleton
class DockerCmdWrapper:
    """docker command wrapper for managing docker containers.
    This class is a singleton, so only one instance will be created.
    Args:
        - run_with_sudo (bool, optional) : run all commands with sudo.
        Default is False.
    """

    def __init__(self, run_with_sudo: bool = True):
        if not hasattr(self, "_initialized"):
            self._initialized = True
        self._run_with_sudo = run_with_sudo

    def run_cmd(self, cmd: str):
        """Run docker command.
        Args:
            - cmd (str) : docker command to run
        """
        if self._run_with_sudo:
            cmd = f"sudo {cmd}"
        return subprocess.run(cmd, shell=True, check=True)

    def check_container(self, container: str):
        """Check if container exists.

        Args:
            container (str): container name or id
            _run_with_sudo (bool, optional): run command with sudo.
            Defaults to None.

        Raises:
            e: subprocess.CalledProcessError if container not found
        """

        cmd = f"docker inspect {container}"
        try:
            exec_cmd(cmd, self._run_with_sudo, bash=False)
        except subprocess.CalledProcessError as e:
            if not self._run_with_sudo:
                raise ContainerNotFoundError(
                    f"Container {container} not found. "
                    f"Try to set _run_with_sudo=True."
                ) from e
            raise ContainerNotFoundError(
                f"Container {container} not found."
            ) from e

    def is_container_exist(self, container: str) -> bool:
        """Check if container exists.

        Args:
            container (str): container name or id

        Returns:
            bool: True if container exists, False otherwise
        """
        cmd = f"docker inspect {container}"
        try:
            exec_cmd(cmd, self._run_with_sudo, bash=False)
        except subprocess.CalledProcessError:
            return False
        return True

    def get_container(self, prefix: Optional[str] = None) -> list[str]:
        """Get all containers with given prefix.

        Args:
            - prefix (str, optional): container name prefix. Default is None.
        """
        cmd = "docker ps --format '{{.Names}}'"
        if prefix:
            cmd += f" | grep {prefix}"
        return (
            exec_cmd(cmd, self._run_with_sudo, bash=False, stdout=True)
            .stdout.decode()
            .splitlines()
        )



import multiprocessing as mp
import time

from contcfg.container_net_ctrl import ConNetServer, ConNetController
from contcfg.container_net_ctrl import CtrlMsg, CtrlAction


def test_net_ctrl_comm():
    # 1. Create a server
    server = ConNetServer(
        min_rate=1,
        max_rate=10,
        interval=5,
        prefix="test",
        rate_unit="mbit",
        interval_unit="s",
        _run_with_sudo=True,
    )
    task = mp.Process(target=server.start)
    task.start()

    time.sleep(1)
    # 2. Create a client
    sender = ConNetController(_run_with_sudo=True)

    for c in sender.find_all_containers("constellation-test-worker-"):
        sender.add_container(c)
        time.sleep(0.1)

    # 4. Stop
    sender.stop_server()
    task.join()
    print("Test finished.")


if __name__ == "__main__":
    test_net_ctrl_comm()



from .msg import CtrlMsg, CtrlAction
from .container_net_server import ConNetServer
from .container_net_controller import ConNetController

__all__ = ["CtrlMsg", "CtrlAction", "ConNetServer", "ConNetController"]



from pathlib import Path
from typing import Optional
import subprocess


def get_script(script_name: str) -> Path:
    scripts_path = Path(__file__).resolve().parent.parent / "utils" / "scripts"
    assert scripts_path.exists(), f"Scripts path {scripts_path} not found"
    return scripts_path / script_name


def check_scripts():
    script_names = [
        "clear_tc_rules.sh",
        "docker_tcconfig.sh",
        "find_container_ip.sh",
        "find_container_veth_name.sh",
        "set_network_limit.sh",
    ]
    for script_name in script_names:
        assert get_script(
            script_name
        ).exists(), f"Script {script_name} not found"


def singleton(cls):
    """Singleton decorator for classes.
    Note: This decorator will call __init__ method
    every time the instance is created.
    """
    instances = {}

    def get_instance(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        instances[cls].__init__(*args, **kwargs)
        return instances[cls]

    return get_instance


def exec_cmd(
    cmd: str,
    run_with_sudo: Optional[bool] = None,
    bash: bool = True,
    stdout=False,
) -> subprocess.CompletedProcess:
    """Execute a command.

    Args:
        - cmd (str): command to execute
        - _run_with_sudo (bool, optional): run command with sudo.
        Defaults to None.
        - bash (bool, optional): run command with bash.
        Defaults to True.
    """

    if bash and not cmd.strip().startswith("bash"):
        cmd = f"bash {cmd}"
    if run_with_sudo:
        cmd = f"sudo {cmd}"
    if stdout:
        output = subprocess.PIPE
    else:
        output = subprocess.DEVNULL
    return subprocess.run(
        cmd,
        shell=True,
        check=True,
        stdout=output,
    )



from .base import check_scripts, get_script, singleton, exec_cmd
from .tc_base import TC_BANDWIDTH_UNITS, split_raw_str_rate
from .dockercmd_wrapper import DockerCmdWrapper
from .exception import RateValueError

from typing import Union


@singleton
class TCCmdWrapper:
    """tc command wrapper for setting bandwidth
    limits between docker containers.
    This class is a singleton,
    so only one instance will be created.
    Args:
        - run_with_sudo (bool, optional) : run all commands with sudo.
        Default is False.
    """

    def __init__(self, run_with_sudo: bool = False):
        if not hasattr(self, "_initialized"):
            check_scripts()  # check if all scripts are present
            self._exec_script = get_script("docker_tcconfig.sh")
            self._initialized = True

        self._run_with_sudo = run_with_sudo

    def set_bandwidth(
        self,
        container1: str,
        container2: str,
        bandwidth: Union[int, str],
        bandwidth_unit: str = "mbit",
        _run_with_sudo: bool = False,
    ):
        """Set bandwidth limit between two containers.
        Args:
            - container1 (str) : container name or id
            - container2 (str) : container name or id
            - bandwidth (int) : bandwidth limit
            - bandwidth_unit (str, optional) : bandwidth unit.
            Default is "mbit".
            - _run_with_sudo (bool, optional) : run command with sudo.
            Default is None.
        """
        if isinstance(bandwidth, str):
            bandwidth, bandwidth_unit = split_raw_str_rate(bandwidth)
        run_with_sudo = self._run_with_sudo or _run_with_sudo
        _check_container = DockerCmdWrapper(
            run_with_sudo
        ).check_container  # function call to check if container exists

        _check_container(container1)
        _check_container(container2)
        self._check_bandwidth(bandwidth, bandwidth_unit)
        cmd = (
            f"{self._exec_script} {container1} {container2} "
            + f"{bandwidth}{bandwidth_unit}"
        )
        exec_cmd(cmd, run_with_sudo)

    def init_htb(self, container: str, _run_with_sudo: bool = False):
        """Initialize htb qdisc for container.
        Args:
            - container (str) : container name or id
            - _run_with_sudo (bool, optional) : run command with sudo.
            Default is None.
        """
        run_with_sudo = self._run_with_sudo or _run_with_sudo
        DockerCmdWrapper(run_with_sudo).check_container(container)
        cmd = f"{self._exec_script} -init {container}"
        exec_cmd(cmd, run_with_sudo)

    def clear_one_container(self, container: str, _run_with_sudo: bool = False):
        """Clear bandwidth limit for one container.
        Args:
            - container (str) : container name or id
            - _run_with_sudo (bool, optional) : run command with sudo.
            Default is None.
        """
        run_with_sudo = self._run_with_sudo or _run_with_sudo
        if DockerCmdWrapper(run_with_sudo).is_container_exist(container):
            cmd = f"{self._exec_script} -c {container}"
            exec_cmd(cmd, run_with_sudo)

    def _check_bandwidth(self, bandwidth: int, bandwidth_unit: str):
        if bandwidth_unit not in TC_BANDWIDTH_UNITS:
            raise RateValueError(f"Invalid bandwidth unit {bandwidth_unit}")
        if bandwidth < 0:
            raise RateValueError(f"Invalid bandwidth {bandwidth}")



from contcfg import TCCmdWrapper
from contcfg.cmd_wrapper import DockerCmdWrapper

if __name__ == "__main__":
    tc = TCCmdWrapper(run_with_sudo=True)
    cmder = DockerCmdWrapper(run_with_sudo=True)
    cmder.check_container("constellation-test-worker-2")
    print(cmder.get_container("constellation-test-worker"))



import threading
import asyncio
from asyncio import Queue

from contcfg.container_net_ctrl.comm import NetCtrlCommServer, NetCtrlCommClient
from contcfg.container_net_ctrl import CtrlMsg, CtrlAction


q = Queue()
socket_path = "/tmp/net_ctrl_socket.sock"


def server():
    server = NetCtrlCommServer(socket_path)
    asyncio.run(server.start(q))


def test_net_ctrl_comm():
    import time

    server_thread = threading.Thread(target=server)
    server_thread.start()

    time.sleep(0.2)

    client = NetCtrlCommClient(socket_path)

    for i in range(19):
        msg = CtrlMsg(action=CtrlAction.SET_BANDWIDTH, container="container1")
        client.send(msg)
        time.sleep(0.1)

    client.stop()

    server_thread.join()


if __name__ == "__main__":
    test_net_ctrl_comm()
    assert q.qsize() == 20



from .tccmd_wrapper import TCCmdWrapper, DockerCmdWrapper
from .exception import RateValueError, ContainerNotFoundError
from .tc_base import split_raw_str_rate

__all__ = [
    "TCCmdWrapper",
    "DockerCmdWrapper",
    "RateValueError",
    "ContainerNotFoundError",
    "split_raw_str_rate",
]
